from typing import Generic, Iterable, TypeVar

from rainbowadn.core import *
from .tlparametres import TLParametres

__all__ = ('TLNode', 'TLNodeFactory',)

ElementType = TypeVar('ElementType')


class TLNode(RecursiveMentionable, Generic[ElementType]):
    def __init__(
            self,
            source: bytes,
            parametres: TLParametres[ElementType],
            node_cache: tuple[MetaOrigin['TLNode[ElementType]'], ...],
            element_cache: tuple[MetaOrigin[ElementType], ...],
    ):
        assert isinstance(source, bytes)
        assert isinstance(parametres, TLParametres)
        assert isinstance(node_cache, tuple)
        assert isinstance(element_cache, tuple)
        self.parametres = parametres
        self.source = source
        if parametres.leaf:
            assert len(node_cache) == 0
            self.elements = parametres.branching
            assert len(element_cache) == self.elements
            self.element_cache = element_cache
        else:
            assert len(element_cache) == 0
            self.nodes = parametres.branching
            assert len(node_cache) == self.nodes
            self.node_cache = node_cache

    def bytes_no(self, index: int) -> bytes:
        assert isinstance(index, int)
        assert index < self.parametres.branching
        return self.source[HashPoint.HASH_LENGTH * index:HashPoint.HASH_LENGTH * (index + 1)]

    def element_no(self, index: int) -> HashPoint[ElementType]:
        assert isinstance(index, int)
        assert index < self.elements
        meta_origin: MetaOrigin[ElementType] = self.element_cache[index]
        return meta_origin.hash_point(self.parametres.tlr.factory, self.bytes_no(index))

    def node_no(self, index: int) -> HashPoint['TLNode[ElementType]']:
        assert isinstance(index, int)
        assert index < self.nodes
        meta_origin: MetaOrigin[TLNode[ElementType]] = self.node_cache[index]
        return meta_origin.hash_point(
            TLNodeFactory(
                self.parametres.params_no(index)
            ),
            self.bytes_no(index)
        )

    async def node_no_resolved(self, index: int) -> 'TLNode[ElementType]':
        assert isinstance(index, int)
        assert index < self.nodes
        return await self.node_no(index).resolve()

    def points(self) -> Iterable[HashPoint]:
        if self.parametres.leaf:
            return [self.element_no(element) for element in range(self.elements)]
        else:
            return [self.node_no(node) for node in range(self.nodes)]

    def __topology_hash__(self) -> bytes:
        return HashPoint.hash(self.source)

    def __bytes__(self):
        return self.source

    def __factory__(self) -> RainbowFactory['TLNode[ElementType]']:
        return TLNodeFactory(
            self.parametres
        )

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        if self.parametres.size == 0:
            return f'-'
        else:
            return f'{tabulate(tab)}'.join(await gather(*(hash_point_format(point, tab) for point in self.points())))

    def unit(self, element: HashPoint[ElementType]) -> 'TLNode[ElementType]':
        assert isinstance(element, HashPoint)
        return TLNode(
            bytes(element),
            self.parametres.unit(),
            (),
            (LocalMetaOrigin(element.origin),),
        )

    def source_without_last(self) -> bytes:
        return self.source[:-HashPoint.HASH_LENGTH]

    async def last(self) -> 'TLNode[ElementType]':
        assert not self.parametres.leaf
        return await self.node_no_resolved(self.nodes - 1)

    async def add(self, element: HashPoint[ElementType]) -> 'TLNode[ElementType]':
        assert isinstance(element, HashPoint)
        if self.parametres.full:
            self_hp = HashPoint.of(self)
            assert isinstance(self_hp, HashPoint)
            unit = self.unit(element)
            assert isinstance(unit, TLNode)
            unit_hp = HashPoint.of(unit)
            assert isinstance(unit_hp, HashPoint)
            return TLNode(
                bytes(self_hp) + bytes(unit_hp),
                self.parametres.superparams(),
                (LocalMetaOrigin(self_hp.origin), LocalMetaOrigin(unit_hp.origin)),
                (),
            )
        elif self.parametres.leaf:
            return TLNode(
                self.source + bytes(element),
                self.parametres.superparams(),
                (),
                self.element_cache + (LocalMetaOrigin(element.origin),),
            )
        elif self.parametres.lastfull:
            unit = self.unit(element)
            assert isinstance(unit, TLNode)
            unit_hp = HashPoint.of(unit)
            assert isinstance(unit_hp, HashPoint)
            return TLNode(
                self.source + bytes(unit_hp),
                self.parametres.superparams(),
                self.node_cache + (LocalMetaOrigin(unit_hp.origin),),
                (),
            )
        else:
            last = await (await self.last()).add(element)
            assert isinstance(last, TLNode)
            last_hp = HashPoint.of(last)
            assert isinstance(last_hp, HashPoint)
            return TLNode(
                self.source_without_last() + bytes(last_hp),
                self.parametres.superparams(),
                self.node_cache[:-1] + (LocalMetaOrigin(last_hp.origin),),
                (),
            )


class TLNodeFactory(RainbowFactory[TLNode[ElementType]], Generic[ElementType]):
    def __init__(
            self,
            parametres: TLParametres[ElementType],
    ):
        assert isinstance(parametres, TLParametres)
        self.parametres = parametres
        if self.parametres.leaf:
            self.nodes, self.elements = 0, parametres.branching
        else:
            self.nodes, self.elements = parametres.branching, 0
        assert isinstance(self.nodes, int)
        assert isinstance(self.elements, int)

    def from_bytes(self, source: bytes, resolver: HashResolver) -> TLNode[ElementType]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return TLNode(
            source,
            self.parametres,
            (ResolverMetaOrigin(resolver),) * self.nodes,
            (ResolverMetaOrigin(resolver),) * self.elements,
        )
