from typing import Generic, Iterable, TypeVar

from rainbowadn.atomic import *
from rainbowadn.core import *
from .tlnode import *
from .tlparametres import TLParametres
from .tlrparametres import TLRParametres

__all__ = ('TLRoot', 'TLRootFactory',)

ElementType = TypeVar('ElementType')


class TLRoot(RecursiveMentionable, Generic[ElementType]):
    def __init__(
            self,
            node: HashPoint[TLNode[ElementType]],
            parametres: TLParametres
    ):
        assert isinstance(node, HashPoint)
        assert isinstance(parametres, TLParametres)
        self.node = node
        self.parametres = parametres

    def points(self) -> Iterable[HashPoint]:
        return [self.node]

    def __bytes__(self):
        return bytes(self.node) + bytes(Integer(self.parametres.size))

    def __factory__(self) -> RainbowFactory['TLRoot[ElementType]']:
        return TLRootFactory(self.parametres.tlr)

    async def node_resolved(self) -> TLNode[ElementType]:
        return await self.node.resolve()

    async def add(self, element: HashPoint[ElementType]) -> 'TLRoot[ElementType]':
        assert isinstance(element, HashPoint)
        node = await (await self.node_resolved()).add(element)
        return TLRoot(
            HashPoint.of(node),
            node.parametres
        )

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return await hash_point_format(self.node, tab)


class TLRootFactory(RainbowFactory[TLRoot[ElementType]]):
    def __init__(
            self,
            tlr: TLRParametres,
    ):
        assert isinstance(tlr, TLRParametres)
        self.tlr = tlr

    def from_bytes(self, source: bytes, resolver: HashResolver) -> TLRoot[ElementType]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        size: int = Integer.from_bytes(source[HashPoint.HASH_LENGTH:], resolver).integer
        assert isinstance(size, int)
        parametres = TLParametres(self.tlr, size)
        return TLRoot(
            ResolverOrigin(
                TLNodeFactory(parametres), source[:HashPoint.HASH_LENGTH], resolver
            ).hash_point(),
            parametres
        )

    def empty(self) -> TLRoot[ElementType]:
        parametres = TLParametres(self.tlr, 0)
        return TLRoot(
            HashPoint.of(TLNode(b'', parametres, (), ())),
            parametres
        )
